/*
 * Copyright (c) 2013 Claus Assmann
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the license/LICENSE.3C file which can be found at the
 * top level of this source code distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: matchhostname.c,v 1.6 2013/10/06 01:04:55 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/string.h"
#include "sm/str.h"
#include "sm/misc.h"

/*
**  MTA_MATCH_HOSTNAME -- try to match hostname against a "cert" pattern
**  Need a better name (more descriptive)!!!
**
**	Parameters:
**		dnsname -- subjectAltName DNS value
**		name -- try to match this name
**
**	Returns:
**		>0: "length" of match (higher: "better")
**		<0: error: SM_E_NOTFOUND: no match
*/

int
mta_match_hostname(sm_str_P dnsname, sm_str_P name)
{
	uint idx;
	size_t nl, dl;

	SM_REQUIRE(NULL != dnsname);
	SM_REQUIRE(NULL != name);

	/* identical? */
	if (sm_str_casecmp(dnsname, name) == 0)
		return (int) sm_str_getlen(name);

	/*
	 * .name requests subdomain match:
	 * .name matches anything.name, i.e., .name must be identical to the
	 * "end" of dnsname (including the "." which is a delimiter!)
	 */

	if ((nl = sm_str_getlen(name)) >= 2
		&& sm_str_rd_elem(name, 0) == '.'
		&& (dl = sm_str_getlen(dnsname)) > nl
		&& sm_str_casecmp_off(name, 0, dnsname, dl - nl) == 0)
	{
		return (int) nl;
	}

	/* is the pattern "*.more" but not just "*.TLD"? */
	if (sm_str_getlen(dnsname) < 4
		|| sm_str_rd_elem(dnsname, 0) != '*'
		|| sm_str_rd_elem(dnsname, 1) != '.'
		|| !sm_is_success(sm_str_f1chr(dnsname, '.', 2, &idx)))
	{
		return sm_err_perm(SM_E_NOTFOUND);
	}

	/* cut off the first subdomain (* matches only one subdomain here!) */
	if (!sm_is_success(sm_str_f1chr(name, '.', 0, &idx))
		|| idx + 1 >= sm_str_getlen(name))
	{
		return sm_err_perm(SM_E_NOTFOUND);
	}
	if (sm_str_casecmp_off(name, idx, dnsname, 1) == 0)
		return (int) sm_str_getlen(dnsname);
	return sm_err_perm(SM_E_NOTFOUND);
}

/*
**  MTA_BEST_MATCH -- try to find best match for hostname against a "cert" pattern
**
**	Parameters:
**		strs -- array of subjectAltName DNS values
**		nstr -- size of strs array
**		name -- try to match this name
**		first -- return first match
**
**	Returns:
**		>=0: index in strs[] that matches (best)
**		<0: error (SM_E_NOTFOUND)
*/

sm_ret_T
mta_best_match(sm_str_P strs[], int nstr, sm_str_P name, bool first)
{
	int r, i, best, best_idx;

	best = 0;
	for (i = 0; i < nstr; ++i) {
		if (!SM_IS_EMPTY_STR(strs[i])) {
			/* try to find "best" match */
			r = mta_match_hostname(strs[i], name);
			if (r > best) {
				if (first) return i;
				best = r;
				best_idx = i;
			}
		}
	}
	if (best > 0)
		return best_idx;
	return sm_err_perm(SM_E_NOTFOUND);
}
